﻿// Copyright (c) Brock Allen & Dominick Baier. All rights reserved.
// Licensed under the Apache License, Version 2.0. See LICENSE in the project root for license information.


using System;
using System.Linq;
using System.Threading.Tasks;
using IdentityModel;
using HaBook.IES.IdentityServer.AzureTableStorage.Entities;
using HaBook.IES.IdentityServer.AzureTableStorage.Interfaces;
using IdentityServer4.Models;
using IdentityServer4.Stores;
using IdentityServer4.Stores.Serialization;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using HaBook.IES.IdentityServer.AzureTableStorage.Options;
using Microsoft.WindowsAzure.Storage.Table;
using Microsoft.WindowsAzure.Storage;

namespace HaBook.IES.IdentityServer.AzureTableStorage.Stores
{
    /// <summary>
    /// Implementation of IDeviceFlowStore thats uses EF.
    /// </summary>
    /// <seealso cref="IdentityServer4.Stores.IDeviceFlowStore" />
    public class DeviceFlowStore : IDeviceFlowStore
    {
       // private readonly IPersistedGrantDbContext _context;
        private readonly IPersistentGrantSerializer _serializer;
        private readonly ILogger _logger;
        private readonly string _connectionString;
        private CloudTable _cloudTable;

        /// <summary>
        /// Initializes a new instance of the <see cref="DeviceFlowStore"/> class.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <param name="serializer">The serializer</param>
        /// <param name="logger">The logger.</param>
        public DeviceFlowStore(
            IOptions<DatabaseOptions> optionsAccessor,
            // IPersistedGrantDbContext context, 
            IPersistentGrantSerializer serializer, 
            ILogger<DeviceFlowStore> logger)
        {
          //  _context = context;
            _serializer = serializer;
            _connectionString = optionsAccessor.Value.ConnectionString;
            _logger = logger;
        }

        async Task<CloudTable> InitTable()
        {
            if (_cloudTable != null) return _cloudTable;

            var storageAccount = CloudStorageAccount.Parse(_connectionString);
            var tableClient = storageAccount.CreateCloudTableClient();
            _cloudTable = tableClient.GetTableReference("IdentityServer4DeviceFlowStore");
            await _cloudTable.CreateIfNotExistsAsync();
            return _cloudTable;
        }

        /// <summary>
        /// Stores the device authorization request.
        /// </summary>
        /// <param name="deviceCode">The device code.</param>
        /// <param name="userCode">The user code.</param>
        /// <param name="data">The data.</param>
        /// <returns></returns>
        public async Task StoreDeviceAuthorizationAsync(string deviceCode, string userCode, DeviceCode data)
        {
            // TODO处理业务逻辑
            //保存信息
            DeviceFlowCodes deviceFlowCodes = ToEntity(data, deviceCode, userCode);

            _logger.LogDebug("storing persisted grant: {DeviceFlowCodes}", deviceFlowCodes);
            var table = await InitTable();
            var operation = TableOperation.InsertOrReplace(deviceFlowCodes);
            var result = await table.ExecuteAsync(operation);
            // _context.DeviceFlowCodes.Add(ToEntity(data, deviceCode, userCode));
            //_context.SaveChanges();
           // return Task.FromResult(0);
        }

        /// <summary>
        /// Finds device authorization by user code.
        /// </summary>
        /// <param name="userCode">The user code.</param>
        /// <returns></returns>
        public async Task<DeviceCode> FindByUserCodeAsync(string userCode)
        {
            // TODO处理业务逻辑
            //查找

            var table = await InitTable();
            Entities.DeviceFlowCodes deviceFlowCodes = null;
            var query = new TableQuery<Entities.DeviceFlowCodes>().Where(TableQuery.GenerateFilterCondition("UserCode", QueryComparisons.Equal, userCode));
            TableContinuationToken continuationToken = null;

            do
            {
                var result = await table.ExecuteQuerySegmentedAsync(query, continuationToken);

                if (result.Results.Count > 0)
                {
                    deviceFlowCodes = result.Results[0];
                    break;
                }

                continuationToken = result.ContinuationToken;
            } while (continuationToken != null);



            //var deviceFlowCodes = _context.DeviceFlowCodes.AsNoTracking().FirstOrDefault(x => x.UserCode == userCode);
            var  model = ToModel(deviceFlowCodes?.Data);

            _logger.LogDebug("{userCode} found in database: {userCodeFound}", userCode, model != null);

            return await Task.FromResult(model);
        }

        /// <summary>
        /// Finds device authorization by device code.
        /// </summary>
        /// <param name="deviceCode">The device code.</param>
        /// <returns></returns>
        public async Task<DeviceCode> FindByDeviceCodeAsync(string deviceCode)
        {
            // TODO处理业务逻辑
            //查找


            var table = await InitTable();
            Entities.DeviceFlowCodes deviceFlowCodes = null;
            var query = new TableQuery<Entities.DeviceFlowCodes>().Where(TableQuery.GenerateFilterCondition("DeviceCode", QueryComparisons.Equal, deviceCode));
            TableContinuationToken continuationToken = null;

            do
            {
                var result = await table.ExecuteQuerySegmentedAsync(query, continuationToken);

                if (result.Results.Count > 0)
                {
                    deviceFlowCodes = result.Results[0];
                    break;
                }

                continuationToken = result.ContinuationToken;
            } while (continuationToken != null);


           // var deviceFlowCodes = _context.DeviceFlowCodes.AsNoTracking().FirstOrDefault(x => x.DeviceCode == deviceCode);
            var model = ToModel(deviceFlowCodes?.Data);

            _logger.LogDebug("{deviceCode} found in database: {deviceCodeFound}", deviceCode, model != null);

            return await Task.FromResult(model);
        }

        /// <summary>
        /// Updates device authorization, searching by user code.
        /// </summary>
        /// <param name="userCode">The user code.</param>
        /// <param name="data">The data.</param>
        /// <returns></returns>
        public async Task UpdateByUserCodeAsync(string userCode, DeviceCode data)
        {
            // TODO处理业务逻辑
            //查找
           // var existing = _context.DeviceFlowCodes.SingleOrDefault(x => x.UserCode == userCode);

            var table = await InitTable();
            Entities.DeviceFlowCodes deviceFlowCodes = null;
            var query = new TableQuery<Entities.DeviceFlowCodes>().Where(TableQuery.GenerateFilterCondition("UserCode", QueryComparisons.Equal, userCode));
            TableContinuationToken continuationToken = null;
            do
            {
                var result = await table.ExecuteQuerySegmentedAsync(query, continuationToken);

                if (result.Results.Count > 0)
                {
                    deviceFlowCodes = result.Results[0];
                    break;
                }

                continuationToken = result.ContinuationToken;
            } while (continuationToken != null);

            if (deviceFlowCodes == null)
            {
                _logger.LogError("{userCode} not found in database", userCode);
                throw new InvalidOperationException("Could not update device code");
            }

            var entity = ToEntity(data, deviceFlowCodes.DeviceCode, userCode);
            _logger.LogDebug("{userCode} found in database", userCode);

            deviceFlowCodes.SubjectId = data.Subject?.FindFirst(JwtClaimTypes.Subject).Value;
            deviceFlowCodes.Data = entity.Data;

            try
            {

                _logger.LogDebug("storing persisted grant: {DeviceFlowCodes}", deviceFlowCodes);
                 // table = await InitTable();
                var operation = TableOperation.InsertOrReplace(deviceFlowCodes);
                var result = await table.ExecuteAsync(operation);
              //  _context.SaveChanges();
            }
            catch (DbUpdateConcurrencyException ex)
            {
                _logger.LogWarning("exception updating {userCode} user code in database: {error}", userCode, ex.Message);
            }

          //  return await Task.FromResult(0);
        }

        /// <summary>
        /// Removes the device authorization, searching by device code.
        /// </summary>
        /// <param name="deviceCode">The device code.</param>
        /// <returns></returns>
        public async Task RemoveByDeviceCodeAsync(string deviceCode)
        {
            //var deviceFlowCodes = _context.DeviceFlowCodes.FirstOrDefault(x => x.DeviceCode == deviceCode);

            var table = await InitTable();
            Entities.DeviceFlowCodes deviceFlowCodes = null;
            var query = new TableQuery<Entities.DeviceFlowCodes>().Where(TableQuery.GenerateFilterCondition("DeviceCode", QueryComparisons.Equal, deviceCode));
            TableContinuationToken continuationToken = null;

            do
            {
                var result = await table.ExecuteQuerySegmentedAsync(query, continuationToken);

                if (result.Results.Count > 0)
                {
                    deviceFlowCodes = result.Results[0];
                    break;
                }

                continuationToken = result.ContinuationToken;
            } while (continuationToken != null);

            if (deviceFlowCodes != null)
            {
                _logger.LogDebug("removing {deviceCode} device code from database", deviceCode);

              //  _context.DeviceFlowCodes.Remove(deviceFlowCodes);

                try
                {
                    //   _context.SaveChanges();
                    table = await InitTable();
                    var operation = TableOperation.Delete(deviceFlowCodes);
                    var result = await table.ExecuteAsync(operation);
                }
                catch (DbUpdateConcurrencyException ex)
                {
                    _logger.LogInformation("exception removing {deviceCode} device code from database: {error}", deviceCode, ex.Message);
                }
            }
            else
            {
                _logger.LogDebug("no {deviceCode} device code found in database", deviceCode);
            }

         //   return Task.FromResult(0);
        }

        private DeviceFlowCodes ToEntity(DeviceCode model, string deviceCode, string userCode)
        {
            if (model == null || deviceCode == null || userCode == null) return null;

            return new DeviceFlowCodes
            {
                DeviceCode = deviceCode,
                UserCode = userCode,
                ClientId = model.ClientId,
                SubjectId = model.Subject?.FindFirst(JwtClaimTypes.Subject).Value,
                CreationTime = model.CreationTime,
                Expiration = model.CreationTime.AddSeconds(model.Lifetime),
                Data = _serializer.Serialize(model)
            };
        }

        private DeviceCode ToModel(string entity)
        {
            if (entity == null) return null;

            return _serializer.Deserialize<DeviceCode>(entity);
        }
    }
}