﻿// Copyright (c) Brock Allen & Dominick Baier. All rights reserved.
// Licensed under the Apache License, Version 2.0. See LICENSE in the project root for license information.


using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using HaBook.IES.IdentityServer.AzureTableStorage.Interfaces;
using HaBook.IES.IdentityServer.AzureTableStorage.Mappers;
using IdentityServer4.Models;
using IdentityServer4.Stores;
using Microsoft.Extensions.Logging;
using Microsoft.EntityFrameworkCore;
using Microsoft.WindowsAzure.Storage.Table;
using HaBook.IES.IdentityServer.AzureTableStorage.Options;
using Microsoft.Extensions.Options;
using Microsoft.WindowsAzure.Storage;
using System.Net;

namespace HaBook.IES.IdentityServer.AzureTableStorage.Stores
{
    /// <summary>
    /// Implementation of IPersistedGrantStore thats uses EF.
    /// </summary>
    /// <seealso cref="IdentityServer4.Stores.IPersistedGrantStore" />
    public class PersistedGrantStore : IPersistedGrantStore
    {
        private readonly string _connectionString;
       // private readonly IPersistedGrantDbContext _context;
        private readonly ILogger _logger;
        private CloudTable _cloudTable;

        /// <summary>
        /// Initializes a new instance of the <see cref="PersistedGrantStore"/> class.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <param name="logger">The logger.</param>
        public PersistedGrantStore(IOptions<DatabaseOptions> optionsAccessor, ILogger<PersistedGrantStore> logger)
        {
            _connectionString = optionsAccessor.Value.ConnectionString;
           // _context = context;
            _logger = logger;
        }
        async Task<CloudTable> InitTable()
        {
            if (_cloudTable != null) return _cloudTable;

            var storageAccount = CloudStorageAccount.Parse(_connectionString);
            var tableClient = storageAccount.CreateCloudTableClient();
            _cloudTable = tableClient.GetTableReference("IdentityServer4PersistedGrantStore");
            await _cloudTable.CreateIfNotExistsAsync();

            return _cloudTable;
        }
        /// <summary>
        /// Stores the asynchronous.
        /// </summary>
        /// <param name="token">The token.</param>
        /// <returns></returns>
        public async Task StoreAsync(PersistedGrant token)
        {
            /**
            var existing = _context.PersistedGrants.SingleOrDefault(x => x.Key == token.Key);
            if (existing == null)
            {
                _logger.LogDebug("{persistedGrantKey} not found in database", token.Key);

                var persistedGrant = token.ToEntity();
                _context.PersistedGrants.Add(persistedGrant);
            }
            else
            {
                _logger.LogDebug("{persistedGrantKey} found in database", token.Key);

                token.UpdateEntity(existing);
            }

            try
            {
                _context.SaveChanges();
            }
            catch (DbUpdateConcurrencyException ex)
            {
                _logger.LogWarning("exception updating {persistedGrantKey} persisted grant in database: {error}", token.Key, ex.Message);
            }

            return Task.FromResult(0);
        **/
            var persistedGrant = token.ToEntity();
            _logger.LogDebug("storing persisted grant: {persistedGrant}", persistedGrant);
            var table = await InitTable();
            var operation = TableOperation.InsertOrReplace(persistedGrant);
            var result = await table.ExecuteAsync(operation);
            _logger.LogDebug("stored {persistedGrantKey} with result {result}", token.Key, result.HttpStatusCode);
        }

        async Task<Entities.PersistedGrant> GetEntityAsync(string key)
        {
            var table = await InitTable();
            Entities.PersistedGrant model = null;
            var query = new TableQuery<Entities.PersistedGrant>().Where(TableQuery.GenerateFilterCondition("PartitionKey", QueryComparisons.Equal, WebUtility.UrlEncode(key)));
            TableContinuationToken continuationToken = null;

            do
            {
                var result = await table.ExecuteQuerySegmentedAsync(query, continuationToken);

                if (result.Results.Count > 0)
                {
                    model = result.Results[0];
                    break;
                }

                continuationToken = result.ContinuationToken;
            } while (continuationToken != null);

            _logger.LogDebug("{persistedGrantKey} found in database: {persistedGrantKeyFound}", key, model != null);
            return model;
        }

        /// <summary>
        /// Gets the grant.
        /// </summary>
        /// <param name="key">The key.</param>
        /// <returns></returns>
        public async Task<PersistedGrant> GetAsync(string key)
        {
            //var persistedGrant = _context.PersistedGrants.AsNoTracking().FirstOrDefault(x => x.Key == key);
            // var model = persistedGrant?.ToModel();
            var entity = await GetEntityAsync(key);
            //_logger.LogDebug("{persistedGrantKey} found in database: {persistedGrantKeyFound}", key, model != null);

            return entity.ToModel();
        }

        async Task<IEnumerable<Entities.PersistedGrant>> GetAllEntitiesAsync(string subjectId)
        {
            var filter = TableQuery.GenerateFilterCondition("RowKey", QueryComparisons.Equal, subjectId);
            var persistedGrants = await GetFilteredEntitiesAsync(filter);
            var persistedGrantList = persistedGrants.ToList();

            _logger.LogDebug("{persistedGrantCount} persisted grants found for subjectId {subjectId}", persistedGrantList.Count, subjectId);
            return persistedGrantList;
        }

        async Task<IEnumerable<Entities.PersistedGrant>> GetAllEntitiesAsync(string subjectId, string clientId)
        {
            var subjectFilter = TableQuery.GenerateFilterCondition("RowKey", QueryComparisons.Equal, subjectId);
            var clientFilter = TableQuery.GenerateFilterCondition("ClientId", QueryComparisons.Equal, clientId);
            var combinedFilter = TableQuery.CombineFilters(subjectFilter, TableOperators.And, clientFilter);

            var persistedGrants = await GetFilteredEntitiesAsync(combinedFilter);
            var persistedGrantList = persistedGrants.ToList();

            _logger.LogDebug("{persistedGrantCount} persisted grants found for subjectId {subjectId}, clientId {clientId}", persistedGrantList.Count, subjectId, clientId);
            return persistedGrantList;
        }

        async Task<IEnumerable<Entities.PersistedGrant>> GetAllEntitiesAsync(string subjectId, string clientId, string type)
        {
            var subjectFilter = TableQuery.GenerateFilterCondition("RowKey", QueryComparisons.Equal, subjectId);
            var clientFilter = TableQuery.GenerateFilterCondition("ClientId", QueryComparisons.Equal, clientId);
            var typeFilter = TableQuery.GenerateFilterCondition("Type", QueryComparisons.Equal, type);
            var combinedFilter = TableQuery.CombineFilters(subjectFilter, TableOperators.And, TableQuery.CombineFilters(clientFilter, TableOperators.And, typeFilter));

            var persistedGrants = await GetFilteredEntitiesAsync(combinedFilter);
            var persistedGrantList = persistedGrants.ToList();

            _logger.LogDebug("{persistedGrantCount} persisted grants found for subjectId {subjectId}, clientId {type}, clientId {type}", persistedGrantList.Count, subjectId, clientId, type);
            return persistedGrantList;
        }

        async Task<IEnumerable<Entities.PersistedGrant>> GetFilteredEntitiesAsync(string filter)
        {
            var table = await InitTable();
            var query = new TableQuery<Entities.PersistedGrant>().Where(filter);
            TableContinuationToken continuationToken = null;
            var persistedGrants = new List<Entities.PersistedGrant>();

            do
            {
                var result = await table.ExecuteQuerySegmentedAsync(query, continuationToken);
                persistedGrants.AddRange(result.Results);
                continuationToken = result.ContinuationToken;
            } while (continuationToken != null);

            return persistedGrants;
        }

        /// <summary>
        /// Gets all grants for a given subject id.
        /// </summary>
        /// <param name="subjectId">The subject identifier.</param>
        /// <returns></returns>
        public async Task<IEnumerable<PersistedGrant>> GetAllAsync(string subjectId)
        {
            //var persistedGrants = _context.PersistedGrants.Where(x => x.SubjectId == subjectId).AsNoTracking().ToList();
            //var model = persistedGrants.Select(x => x.ToModel());

            //_logger.LogDebug("{persistedGrantCount} persisted grants found for {subjectId}", persistedGrants.Count, subjectId);

            //return Task.FromResult(model);
            var entities = await GetAllEntitiesAsync(subjectId);
            return entities.Select(e => e.ToModel());
        }

        /// <summary>
        /// Removes the grant by key.
        /// </summary>
        /// <param name="key">The key.</param>
        /// <returns></returns>
        public async Task RemoveAsync(string key)
        {
            //var persistedGrant = _context.PersistedGrants.FirstOrDefault(x => x.Key == key);
            //if (persistedGrant!= null)
            //{
            //    _logger.LogDebug("removing {persistedGrantKey} persisted grant from database", key);

            //    _context.PersistedGrants.Remove(persistedGrant);

            //    try
            //    {
            //        _context.SaveChanges();
            //    }
            //    catch(DbUpdateConcurrencyException ex)
            //    {
            //        _logger.LogInformation("exception removing {persistedGrantKey} persisted grant from database: {error}", key, ex.Message);
            //    }
            //}
            //else
            //{
            //    _logger.LogDebug("no {persistedGrantKey} persisted grant found in database", key);
            //}

            //return Task.FromResult(0);

            var persistedGrant = await GetEntityAsync(key);
            if (persistedGrant == null)
            {
                _logger.LogDebug("no {persistedGrantKey} persisted grant found in database", key);
                return;
            }

            var table = await InitTable();
            var operation = TableOperation.Delete(persistedGrant);
            var result = await table.ExecuteAsync(operation);
            _logger.LogDebug("removed {persistedGrantKey} from database with result {result}", key, result.HttpStatusCode);
        }



        /// <summary>
        /// Removes all grants for a given subject id and client id combination.
        /// </summary>
        /// <param name="subjectId">The subject identifier.</param>
        /// <param name="clientId">The client identifier.</param>
        /// <returns></returns>
        public async Task RemoveAllAsync(string subjectId, string clientId)
        {
            //var persistedGrants = _context.PersistedGrants.Where(x => x.SubjectId == subjectId && x.ClientId == clientId).ToList();

            //_logger.LogDebug("removing {persistedGrantCount} persisted grants from database for subject {subjectId}, clientId {clientId}", persistedGrants.Count, subjectId, clientId);

            //_context.PersistedGrants.RemoveRange(persistedGrants);

            //try
            //{
            //    _context.SaveChanges();
            //}
            //catch (DbUpdateConcurrencyException ex)
            //{
            //    _logger.LogInformation("removing {persistedGrantCount} persisted grants from database for subject {subjectId}, clientId {clientId}: {error}", persistedGrants.Count, subjectId, clientId, ex.Message);
            //}

            //return Task.FromResult(0);


            var persistedGrants = await GetAllEntitiesAsync(subjectId, clientId);
            var persistedGrantList = persistedGrants.ToList();
            _logger.LogDebug("removing {persistedGrantCount} persisted grants from database for subject {subjectId}, clientId {clientId}", persistedGrantList.Count, subjectId, clientId);

            await RemoveAllAsync(persistedGrantList);
        }

        /// <summary>
        /// Removes all grants of a give type for a given subject id and client id combination.
        /// </summary>
        /// <param name="subjectId">The subject identifier.</param>
        /// <param name="clientId">The client identifier.</param>
        /// <param name="type">The type.</param>
        /// <returns></returns>
        public async Task RemoveAllAsync(string subjectId, string clientId, string type)
        {
            /**
            var persistedGrants = _context.PersistedGrants.Where(x =>
                x.SubjectId == subjectId &&
                x.ClientId == clientId &&
                x.Type == type).ToList();

            _logger.LogDebug("removing {persistedGrantCount} persisted grants from database for subject {subjectId}, clientId {clientId}, grantType {persistedGrantType}", persistedGrants.Count, subjectId, clientId, type);

            _context.PersistedGrants.RemoveRange(persistedGrants);

            try
            {
                _context.SaveChanges();
            }
            catch (DbUpdateConcurrencyException ex)
            {
                _logger.LogInformation("exception removing {persistedGrantCount} persisted grants from database for subject {subjectId}, clientId {clientId}, grantType {persistedGrantType}: {error}", persistedGrants.Count, subjectId, clientId, type, ex.Message);
            }

            return Task.FromResult(0);
            **/
            var persistedGrants = await GetAllEntitiesAsync(subjectId, clientId, type);
            var persistedGrantList = persistedGrants.ToList();
            _logger.LogDebug("removing {persistedGrantCount} persisted grants from database for subject {subjectId}, clientId {clientId}, grantType {persistedGrantType}", persistedGrantList.Count, subjectId, clientId, type);

            await RemoveAllAsync(persistedGrantList);
        }


        private async Task RemoveAllAsync(IEnumerable<Entities.PersistedGrant> persistedGrants)
        {
            var table = await InitTable();

            foreach (var persistedGrant in persistedGrants)
            {
                var operation = TableOperation.Delete(persistedGrant);
                var result = await table.ExecuteAsync(operation);
                _logger.LogDebug("removed {persistedGrantKey} from database with result {result}", persistedGrant.PartitionKey,
                    result.HttpStatusCode);
            }
        }
    }
}