﻿// Copyright (c) Brock Allen & Dominick Baier. All rights reserved.
// Licensed under the Apache License, Version 2.0. See LICENSE in the project root for license information.


using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using HaBook.IES.IdentityServer.AzureTableStorage.Options;
using HaBook.IES.IdentityServer.AzureTableStorage.Mappers;
using IdentityServer4.Models;
using IdentityServer4.Stores;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Logging;
using Microsoft.WindowsAzure.Storage.Table;
using Microsoft.Extensions.Options;
using Microsoft.WindowsAzure.Storage;
using Entities=  HaBook.IES.IdentityServer.AzureTableStorage.Entities;
using System.Threading;

namespace HaBook.IES.IdentityServer.AzureTableStorage.Stores
{
    /// <summary>
    /// Implementation of IResourceStore thats uses EF.
    /// </summary>
    /// <seealso cref="IResourceStore" />
    public class ResourceStore : IResourceStore
    {

        
        //private readonly IConfigurationDbContext _context;
        private readonly ILogger<ResourceStore> _logger;
        private readonly string _connectionString;
        private CloudTable _cloudTable;

        /// <summary>
        /// Initializes a new instance of the <see cref="ResourceStore"/> class.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <param name="logger">The logger.</param>
        /// <exception cref="ArgumentNullException">context</exception>
        public ResourceStore(IOptions<DatabaseOptions> optionsAccessor, ILogger<ResourceStore> logger)
        {
            _connectionString = optionsAccessor.Value.ConnectionString;
           // _context = context ?? throw new ArgumentNullException(nameof(context));
            _logger = logger;
        }

        async Task<CloudTable> InitTable()
        {
            if (_cloudTable != null) return _cloudTable;

            var storageAccount = CloudStorageAccount.Parse(_connectionString);
            var tableClient = storageAccount.CreateCloudTableClient();
            _cloudTable = tableClient.GetTableReference("IdentityServer4ApiResourceStore");
            await _cloudTable.CreateIfNotExistsAsync();

            return _cloudTable;
        }


        /// <summary>
        /// Finds the API resource by name.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns></returns>
        public async Task<ApiResource> FindApiResourceAsync(string name)
        {
            //处理实体业务
            /*
            var query =
                from apiResource in _context.ApiResources
                where apiResource.Name == name
                select apiResource;

            var apis = query
                .Include(x => x.Secrets)
                .Include(x => x.Scopes)
                    .ThenInclude(s => s.UserClaims)
                .Include(x => x.UserClaims)
                .Include(x => x.Properties)
                .AsNoTracking();

            var api = apis.FirstOrDefault();

            if (api != null)
            {
                _logger.LogDebug("Found {api} API resource in database", name);
            }
            else
            {
                _logger.LogDebug("Did not find {api} API resource in database", name);
            }

            return Task.FromResult(api.ToModel());
            */

            var table = await InitTable();
            Entities.ApiResource api = null;
            var query = new TableQuery<Entities.ApiResource>().Where(TableQuery.GenerateFilterCondition("Name", QueryComparisons.Equal,name));
            TableContinuationToken continuationToken = null;

            do
            {
                var result = await table.ExecuteQuerySegmentedAsync(query, continuationToken);

                if (result.Results.Count > 0)
                {
                    api = result.Results[0];
                    break;
                }

                continuationToken = result.ContinuationToken;
            } while (continuationToken != null);

            return await  Task.FromResult(api.ToModel());

        }

        /// <summary>
        /// Gets API resources by scope name.
        /// </summary>
        /// <param name="scopeNames"></param>
        /// <returns></returns>
        public async  Task<IEnumerable<ApiResource>> FindApiResourcesByScopeAsync(IEnumerable<string> scopeNames)
        {
            /*

            var names = scopeNames.ToArray();

            var query =
                from api in _context.ApiResources
                where api.Scopes.Where(x=>names.Contains(x.Name)).Any()
                select api;

            var apis = query
                .Include(x => x.Secrets)
                .Include(x => x.Scopes)
                    .ThenInclude(s => s.UserClaims)
                .Include(x => x.UserClaims)
                .Include(x => x.Properties)
                .AsNoTracking();

            var results = apis.ToArray();
            var models = results.Select(x => x.ToModel()).ToArray();

            _logger.LogDebug("Found {scopes} API scopes in database", models.SelectMany(x => x.Scopes).Select(x => x.Name));

            return Task.FromResult(models.AsEnumerable());
            */
            CloudStorageAccount storageAccount = CloudStorageAccount.Parse(_connectionString);
            CloudTableClient tableClient = storageAccount.CreateCloudTableClient();
            CloudTable table = tableClient.GetTableReference("IdentityServer4ApiResourceStore");
            CloudTable scope_table = tableClient.GetTableReference("IdentityServer4ApiScopeStore");
            Action<IList<Entities.ApiScope>> onProgress = null;
            CancellationToken ct = default(CancellationToken);
            var scopes = new List<Entities.ApiScope>();
            TableContinuationToken token = null;
            foreach (string name in scopeNames) {
                do
                {
                    var query = new TableQuery<Entities.ApiScope>().Where(TableQuery.GenerateFilterCondition("Name", QueryComparisons.Equal, name));
                    TableQuerySegment<Entities.ApiScope> seg = await  scope_table.ExecuteQuerySegmentedAsync<Entities.ApiScope>(query, token);
                    token = seg.ContinuationToken;
                    scopes.AddRange(seg);
                    onProgress?.Invoke(scopes);

                } while (token != null && !ct.IsCancellationRequested);
            }
            var items = new List<Entities.ApiResource>();
            TableContinuationToken sctoken = null;
            Action<IList<Entities.ApiResource>> sconProgress = null;
            foreach (Entities.ApiScope scope in scopes)
            {
                do
                {
                    var query = new TableQuery<Entities.ApiResource>().Where(TableQuery.GenerateFilterCondition("Id", QueryComparisons.Equal, scope.ApiResourceId+""));
                    TableQuerySegment<Entities.ApiResource> seg = await scope_table.ExecuteQuerySegmentedAsync<Entities.ApiResource>(query, sctoken);
                    sctoken = seg.ContinuationToken;
                    items.AddRange(seg);
                    sconProgress?.Invoke(items);

                } while (sctoken != null && !ct.IsCancellationRequested);
            }
            var models = items.Select(x => x.ToModel()).ToArray();
            return await Task.FromResult(models.AsEnumerable());
            }
      

        /// <summary>
        /// Gets identity resources by scope name.
        /// </summary>
        /// <param name="scopeNames"></param>
        /// <returns></returns>
        public async Task<IEnumerable<IdentityResource>> FindIdentityResourcesByScopeAsync(IEnumerable<string> scopeNames)
        {
            var scopes = scopeNames.ToArray();
            CloudStorageAccount storageAccount = CloudStorageAccount.Parse(_connectionString);
            CloudTableClient tableClient = storageAccount.CreateCloudTableClient();
            CloudTable table = tableClient.GetTableReference("IdentityServer4IdentityResourceStore");
            //var query =
            //        from identityResource in _context.IdentityResources
            //        where scopes.Contains(identityResource.Name)
            //        select identityResource;

            //var resources = query
            //        .Include(x => x.UserClaims)
            //        .Include(x => x.Properties)
            //        .AsNoTracking();
            TableQuery<Entities.IdentityResource> query = new TableQuery<Entities.IdentityResource>();
            TableContinuationToken continuationToken = null;
            List<Entities.IdentityResource> resources = null;
            do
            {
                var result = await table.ExecuteQuerySegmentedAsync(query, continuationToken);
                if (result.Results.Count > 0)
                {
                        resources = result.ToList();
                    break;
                }
                continuationToken = result.ContinuationToken;
            } while (continuationToken != null);
            resources.Where<Entities.IdentityResource>(x => scopeNames.Contains<string>(x.Name) );
            var results = resources.ToArray();
            _logger.LogDebug("Found {scopes} identity scopes in database", results.Select(x => x.Name));
            return await Task.FromResult(results.Select(x => x.ToModel()).ToArray().AsEnumerable());
        }

        /// <summary>
        /// Gets all resources.
        /// </summary>
        /// <returns></returns>
        public async Task<Resources> GetAllResourcesAsync()
        {
            CloudStorageAccount storageAccount = CloudStorageAccount.Parse(_connectionString);
            CloudTableClient tableClient = storageAccount.CreateCloudTableClient();
            CloudTable table = tableClient.GetTableReference("IdentityServer4IdentityResourceStore");

            TableQuery<Entities.IdentityResource> query = new TableQuery<Entities.IdentityResource>();
            TableContinuationToken continuationToken = null;
            List<Entities.IdentityResource> identity = null;
            TableQuerySegment<Entities.IdentityResource> result = null;
            do
            {
                result = await table.ExecuteQuerySegmentedAsync(query, continuationToken);
                if (result.Results.Count > 0)
                {
                    identity = result.ToList();
                    break;
                }
                continuationToken = result.ContinuationToken;
            } while (continuationToken != null);


             
            CloudTable api_table = tableClient.GetTableReference("IdentityServer4ApiResourcesStore");

            TableQuery<Entities.ApiResource> api_query = new TableQuery<Entities.ApiResource>();
            TableContinuationToken api_continuationToken = null;
            List<Entities.ApiResource> apis = null;
            TableQuerySegment<Entities.ApiResource> api_result = null;
            do
            {
                api_result = await api_table.ExecuteQuerySegmentedAsync(api_query, api_continuationToken);
                if (api_result.Results.Count > 0)
                {
                    apis = api_result.ToList();
                    break;
                }
                continuationToken = result.ContinuationToken;
            } while (continuationToken != null);




            //var identity = _context.IdentityResources
            //  .Include(x => x.UserClaims);

            //var apis = _context.ApiResources
            //    .Include(x => x.Secrets)
            //    .Include(x => x.Scopes)
            //        .ThenInclude(s => s.UserClaims)
            //    .Include(x => x.UserClaims)
            //    .Include(x => x.Properties)
            //    .AsNoTracking();
            var resources = new Resources(
                identity.ToArray().Select(x => x.ToModel()).AsEnumerable(),
                apis.ToArray().Select(x => x.ToModel()).AsEnumerable());
            _logger.LogDebug("Found {scopes} as all scopes in database", resources.IdentityResources.Select(x=>x.Name).Union(resources.ApiResources.SelectMany(x=>x.Scopes).Select(x=>x.Name)));
            return await Task.FromResult(resources);
        }
    }
}