﻿// Copyright (c) Brock Allen & Dominick Baier. All rights reserved.
// Licensed under the Apache License, Version 2.0. See LICENSE in the project root for license information.


using System;
using System.Threading.Tasks;
using IdentityServer4.Services;
using System.Linq;
using HaBook.IES.IdentityServer.AzureTableStorage.Interfaces;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.EntityFrameworkCore;
using Microsoft.WindowsAzure.Storage.Table;
using HaBook.IES.IdentityServer.AzureTableStorage.Options;
using Microsoft.Extensions.Options;
using Microsoft.WindowsAzure.Storage;
using HaBook.IES.IdentityServer.AzureTableStorage.Entities;
using System.Collections.Generic;

namespace HaBook.IES.IdentityServer.AzureTableStorage.Services
{
    /// <summary>
    /// Implementation of ICorsPolicyService that consults the client configuration in the database for allowed CORS origins.
    /// </summary>
    /// <seealso cref="IdentityServer4.Services.ICorsPolicyService" />
    public class CorsPolicyService : ICorsPolicyService
    {
        //private readonly IHttpContextAccessor _context;
        private readonly ILogger<CorsPolicyService> _logger;
        private readonly string _connectionString;
        private CloudTable _cloudTable;
        private CloudTable client_cloudTable;
        /// <summary>
        /// Initializes a new instance of the <see cref="CorsPolicyService"/> class.
        /// </summary>
        /// <param name="context">The context.</param>
        /// <param name="logger">The logger.</param>
        /// <exception cref="ArgumentNullException">context</exception>
        public CorsPolicyService(IOptions<DatabaseOptions> optionsAccessor, ILogger<CorsPolicyService> logger)
        {
            _connectionString = optionsAccessor.Value.ConnectionString;
            //_context = context ?? throw new ArgumentNullException(nameof(context));
            _logger = logger;
        }
        async Task<CloudTable> InitTable()
        {
            if (_cloudTable != null) return _cloudTable;
            var storageAccount = CloudStorageAccount.Parse(_connectionString);
            var tableClient = storageAccount.CreateCloudTableClient();
            _cloudTable = tableClient.GetTableReference("IdentityServer4ClientCorsOriginStore");
            await _cloudTable.CreateIfNotExistsAsync();

            return _cloudTable;
        }

        async Task<CloudTable> Client_InitTable()
        {
            if (client_cloudTable != null) return client_cloudTable;
            var storageAccount = CloudStorageAccount.Parse(_connectionString);
            var tableClient = storageAccount.CreateCloudTableClient();
            client_cloudTable = tableClient.GetTableReference("IdentityServer4ClientStore");
            await client_cloudTable.CreateIfNotExistsAsync();

            return client_cloudTable;
        }

        /// <summary>
        /// Determines whether origin is allowed.
        /// </summary>
        /// <param name="origin">The origin.</param>
        /// <returns></returns>
        public async Task<bool> IsOriginAllowedAsync(string origin)
        {

            TableQuery<Client> query = new TableQuery<Client>();
            var table = await Client_InitTable();
            TableContinuationToken continuationToken = null;
            List<Client> origins=null ;
            
            do
            {
                var result = await table.ExecuteQuerySegmentedAsync(query, continuationToken);
                if (result.Results.Count > 0)
                {
                    origins = result.ToList() ;
                    break;
                }
                continuationToken = result.ContinuationToken;
            } while (continuationToken != null);

            // doing this here and not in the ctor because: https://github.com/aspnet/CORS/issues/105
           // var dbContext = _context.HttpContext.RequestServices.GetRequiredService<IConfigurationDbContext>();

            //var origins = dbContext.Clients.SelectMany(x => x.AllowedCorsOrigins.Select(y => y.Origin)).AsNoTracking().ToList();

            //处理子表业务
            var originsStr = origins.SelectMany(x => x.AllowedCorsOrigins.Select(y => y.Origin)).ToList();
            var distinctOrigins = originsStr.Where(x => x != null).Distinct();

            var isAllowed = distinctOrigins.Contains(origin, StringComparer.OrdinalIgnoreCase);

            _logger.LogDebug("Origin {origin} is allowed: {originAllowed}", origin, isAllowed);

            return await Task.FromResult(isAllowed);
        }
    }
}